#include <stdio.h>
#include <pthread.h>
#include <signal.h>
#include <math.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>

#define LOOP_SIZE 1
#define MATRIX_SIZE 300

pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;

typedef struct thread_arg {
  double *matrix;
  int counter;
} s_thread_arg_t, *thread_arg_t;

#define M(i,j) matrix[i+MATRIX_SIZE*j]

double *matrix_init(void);
inline void compute_kernel(double *);
void *thread_main(void *args);
void alarm_handler(int stuff);
int num_threads;

s_thread_arg_t *arguments=NULL;

int start = 0;

/*
 * cpu_sharing <num threads> <time limit>
 * 
 * output number of operations completed by each thread 
 * before the time limit
 *
 */
int main(int argc, char **argv) {
  int i;
  int time_limit;
  pthread_t *threads=NULL;

  if (argc != 3) {
    fprintf(stderr,"Usage: %s <num threads> <time limit>\n",argv[0]);
    exit(-1);
  }
  if ((num_threads = atoi(argv[1])) <= 0) {
    fprintf(stderr,"Invalid number of threads\n");
  }
  if ((time_limit = atoi(argv[2])) <= 0) {
    fprintf(stderr,"Invalid time limit\n");
  }

  /* allocate the array of thread descriptors */
  threads = (pthread_t *)calloc(num_threads,sizeof(pthread_t));

  /* allocate the array of counters */
  arguments = (s_thread_arg_t *)calloc(num_threads,sizeof(s_thread_arg_t));


 fprintf(stderr,"Creating %d threads ",num_threads);
  for (i=0;i<num_threads;i++) {
   fprintf(stderr,".");
   arguments[i].matrix = matrix_init();
   arguments[i].counter = 0;
   if (pthread_create(&(threads[i]),NULL,thread_main,&(arguments[i]))) {
     fprintf(stderr,"Error: pthread_create\n");
     exit(-1);
   }
  }
  fprintf(stderr," done\n");
  
  
//  fprintf(stderr,"Setting up signal!\n");
  signal(SIGALRM, alarm_handler);
  alarm(time_limit);
  
  
//  fprintf(stderr,"Freeing threads!\n");
  pthread_mutex_lock(&mut);
  start = 1;
  pthread_cond_broadcast(&cond);
  pthread_mutex_unlock(&mut);
//  fprintf(stderr,"Freeing threads done!\n");
  
  for (i=0;i<num_threads;i++) {
    pthread_join(threads[i], NULL);
  }
  
  alarm_handler(0);

  return 0;
}

void *thread_main(void *args) {
  thread_arg_t arg = args;

  arg->counter = 0;
  
  pthread_mutex_lock(&mut);
//  fprintf(stderr,"Waiting for a signal!\n");
  while(!start) {
    pthread_cond_wait(&cond, &mut);
  }
//  fprintf(stderr,"Free! Let's go!\n");
  pthread_mutex_unlock(&mut);

//  fprintf(stderr,"Hello!\n");

  while(1) {
    compute_kernel(arg->matrix);
    arg->counter += LOOP_SIZE;
    //    fprintf(stderr,"%d\n",arg->counter);
  }
 
  pthread_exit(NULL);
}

double *matrix_init(void) {
  int i,j;

  double *matrix=calloc(MATRIX_SIZE*MATRIX_SIZE,sizeof(double));

  for (i=0;i<MATRIX_SIZE;i++) {
    for (j=0;j<MATRIX_SIZE;j++) {
      M(i,j) = (double) sqrt(i+12*j);
    }
  }
  return matrix;
}

/*
 *
 */
inline void compute_kernel(double *matrix) {
  int l,i,j;
  double alpha=0.0;
 
  for (l=0;l<LOOP_SIZE;l++) {
    for (i=1;i<MATRIX_SIZE;i++) {
      for (j=1;j<MATRIX_SIZE;j++) {
        if ((i % 3) == 0)
          alpha = alpha + M(i-1,j);
        M(i,j) = M(i-1,j-1) + alpha;
      }
    }
  }
  return;
}

/*
 *
 */
void alarm_handler(int x) {
  int i;
  for (i=0;i<num_threads;i++) {
    fprintf(stdout,"%d\n",arguments[i].counter);
  }
  exit(0);
}
