#include <stdio.h>
#include <pthread.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#define LOOP_SIZE 1
#define MATRIX_SIZE 300


void compute_kernel(double **matrix);
void *thread_main(void *args);
void alarm_handler(int stuff);
int num_threads;
int *counters;

/*
 * cpu_sharing <num threads> <time limit>
 * 
 * output number of operations completed by each thread 
 * before the time limit
 *
 */
int main(int argc, char **argv) {
  int i;
  int time_limit;
  pthread_t *threads=NULL;

  if (argc != 3) {
    fprintf(stderr,"Usage: %s <num threads> <time limit>\n",argv[0]);
    exit(-1);
  }
  if ((num_threads = atoi(argv[1])) <= 0) {
    fprintf(stderr,"Invalid number of threads\n");
  }
  if ((time_limit = atoi(argv[2])) <= 0) {
    fprintf(stderr,"Invalid time limit\n");
  }

  /* allocate the array of thread descriptors */
  threads = (pthread_t *)calloc(num_threads,sizeof(pthread_t));

  /* allocate the array of counters */
  counters = (int *)calloc(num_threads,sizeof(int));


 fprintf(stderr,"Creating %d threads ",num_threads);
  for (i=0;i<num_threads;i++) {
   fprintf(stderr,".");
   if (pthread_create(&(threads[i]),NULL,thread_main,&(counters[i]))) {
     fprintf(stderr,"Error: pthread_create\n");
     exit(-1);
   }
  }
 fprintf(stderr," done\n");


  signal(SIGALRM, alarm_handler);
  alarm(time_limit);

  for (i=0;i<num_threads;i++) {
    pthread_join(threads[i], NULL);
  }

  return 0;
}

void *thread_main(void *args) {
  int *counter = (int*)args;
  double **matrix = NULL;
  int i,j;

  matrix = (double **)calloc(MATRIX_SIZE,sizeof(double*)); 
  for (i=0;i<MATRIX_SIZE;i++) 
    matrix[i] = (double *)calloc(MATRIX_SIZE,sizeof(double));

  for (i=0;i<MATRIX_SIZE;i++) {
    for (j=0;j<MATRIX_SIZE;j++) {
      matrix[i][j] = (double) sqrt(i+12*j);
    }
  }


  *counter = 0;

  while(1) {
    compute_kernel(matrix);
    *counter += LOOP_SIZE;
  }
 
  pthread_exit(NULL);
}

/*
 *
 */
void compute_kernel(double **matrix) {
  int l,i,j;
  double alpha=0.0;
 
  for (l=0;l<LOOP_SIZE;l++) {
    for (i=1;i<MATRIX_SIZE;i++) {
      for (j=1;j<MATRIX_SIZE;j++) {
        if ((i % 3) == 0)
          alpha = alpha + matrix[i-1][j];
        matrix[i][j] = matrix[i-1][j-1] + alpha;
      }
    }
  }
  return;
}

/*
 *
 */
void alarm_handler(int x) {
  int i;
  for (i=0;i<num_threads;i++) {
    fprintf(stdout,"%d\n",counters[i]);
  }
  exit(0);
}
