#include <stdio.h>
#include <pthread.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#define LOOP_SIZE 10
#define MATRIX_SIZE 300

double matrix[MATRIX_SIZE][MATRIX_SIZE];

void matrix_init(void);
void compute_kernel1(void);
void compute_kernel2(int *add);
void *thread_main1(void *args);
void *thread_main2(void *args);
void alarm_handler(int stuff);
int num_threads1,num_threads2;
int *counters1;
int *counters2;

/*
 * cpu_sharing <num threads> <time limit>
 * 
 * output number of operations completed by each thread 
 * before the time limit
 *
 */
int main(int argc, char **argv) {
  int i;
  int time_limit;
  pthread_t *threads1=NULL;
  pthread_t *threads2=NULL;

  if (argc != 4) {
    fprintf(stderr,"Usage: %s <num threads #1> <num threads #2> <time limit>\n",argv[0]);
    exit(-1);
  }
  if ((num_threads1 = atoi(argv[1])) < 0) {
    fprintf(stderr,"Invalid number of threads #1\n");
    exit(1);
  }
  if ((num_threads2 = atoi(argv[2])) < 0) {
    fprintf(stderr,"Invalid number of threads #2\n");
    exit(1);
  }
  if ((time_limit = atoi(argv[3])) <= 0) {
    fprintf(stderr,"Invalid time limit\n");
    exit(1);
  }

  /* allocate the array of thread descriptors */
  if (num_threads1)
    threads1 = (pthread_t *)calloc(num_threads1,sizeof(pthread_t));
  if (num_threads2)
    threads2 = (pthread_t *)calloc(num_threads2,sizeof(pthread_t));

  /* allocate the array of counters */
  if (num_threads1)
    counters1 = (int *)calloc(num_threads1,sizeof(int));
  if (num_threads2)
    counters2 = (int *)calloc(num_threads2,sizeof(int));

  matrix_init();

 fprintf(stderr,"Creating %d threads of type #1 ",num_threads1);
  for (i=0;i<num_threads1;i++) {
   fprintf(stderr,".");
   if (pthread_create(&(threads1[i]),NULL,thread_main1,&(counters1[i]))) {
     fprintf(stderr,"Error: pthread_create\n");
     exit(-1);
   }
  }
 fprintf(stderr," done\n");
 fprintf(stderr,"Creating %d threads of type #2 ",num_threads2);
  for (i=0;i<num_threads2;i++) {
   fprintf(stderr,".");
   if (pthread_create(&(threads2[i]),NULL,thread_main2,&(counters2[i]))) {
     fprintf(stderr,"Error: pthread_create\n");
     exit(-1);
   }
  }
 fprintf(stderr," done\n");


  signal(SIGALRM, alarm_handler);
  alarm(time_limit);

  for (i=0;i<num_threads1;i++) {
    pthread_join(threads1[i], NULL);
  }
  for (i=0;i<num_threads2;i++) {
    pthread_join(threads2[i], NULL);
  }

  return 0;
}

/*
 *
 */
void *thread_main1(void *args) {
  int *counter = (int*)args;

  *counter = 0;

  while(1) {
    compute_kernel1();
    *counter += LOOP_SIZE;
  }
 
  pthread_exit(NULL);
}

/*
 *
 */

void matrix_init(void) {
  int i,j;

  for (i=0;i<MATRIX_SIZE;i++) {
    for (j=0;j<MATRIX_SIZE;j++) {
      matrix[i][j] = (double) sqrt(i+12*j);
    }
  }
}

/*
 *
 */
void *thread_main2(void *args) {
  int *counter = (int*)args;

  *counter = 0;

  while(1) {
    compute_kernel2(counter);
    *counter += 1;
  }
 
  pthread_exit(NULL);
}


/*
 *
 */
void compute_kernel1(void) {
  int l,i,j;
  double alpha=0.0;
 
  for (l=0;l<LOOP_SIZE;l++) {
    for (i=1;i<MATRIX_SIZE;i++) {
      for (j=1;j<MATRIX_SIZE;j++) {
        if ((i % 3) == 0)
          alpha = alpha + matrix[i-1][j];
        matrix[i][j] = matrix[i-1][j-1] + alpha;
      }
    }
  }
  return;
}

/*
 *
 */
void compute_kernel2(int *add) {
  int i;
  FILE *f;
  char filename[128];

  sprintf(filename,"/tmp/bogus_file-%d",(int)add);
  if ((f = fopen(filename,"w")) == NULL) {
     fprintf(stderr,"Cannot open file %s\n",filename);
     return;
   }
   for (i=0;i<2500;i++)
     fprintf(f,"a");
   fclose(f);
  if ((f = fopen(filename,"r")) == NULL) {
     fprintf(stderr,"Cannot open file %s\n",filename);
     return;
   }
   while (fgetc(f) != EOF);
   fclose(f);
  return;
}


/*
 *
 */
void alarm_handler(int x) {
  int i;
  for (i=0;i<num_threads1;i++) {
    fprintf(stdout,"%d\n",counters1[i]);
  }
  for (i=0;i<num_threads2;i++) {
    fprintf(stdout,"%d\n",counters2[i]);
  }
  exit(0);
}
